@@ -12,13 +12,19 @@
 # License; see this file: https://github.com/RDFLib/rdflib/blob/master/LICENSE
 
 from xmptools.xmptools import makeFileURI, isContainerItemPredicate, makeContainerItemPredicate
-from xmptools.xmptools import XMPMetadata, FileTypeError, XMPParser, XMPSerializer
+from xmptools.xmptools import cbd, identity
+from xmptools.xmptools import XMPMetadata
 from xmptools.xmptools import DC, XMP, EXIF, CRS
 from xmptools.xmptools import JPEG_EXTENSIONS, TIFF_EXTENSIONS, RAW_EXTENSIONS, DNG_EXTENSIONS
 from xmptools.xmptools import XMP_EXTENSIONS
+from xmptools.xmptools import FileTypeError, XMPParser, XMPSerializer
 
-__all__ = ['makeFileURI', 'isContainerItemPredicate', 'makeContainerItemPredicate',
-           'XMPMetadata', 'FileTypeError', 'XMPParser', 'XMPSerializer',
-           'DC', 'XMP', 'EXIF', 'CRS',
-           'JPEG_EXTENSIONS', 'TIFF_EXTENSIONS', 'RAW_EXTENSIONS', 'DNG_EXTENSIONS',
-           'XMP_EXTENSIONS']
+__all__ = [
+    'makeFileURI', 'isContainerItemPredicate', 'makeContainerItemPredicate',
+    'cbd', 'identity',
+    'XMPMetadata',
+    'DC', 'XMP', 'EXIF', 'CRS',
+    'JPEG_EXTENSIONS', 'TIFF_EXTENSIONS', 'RAW_EXTENSIONS', 'DNG_EXTENSIONS',
+    'XMP_EXTENSIONS',
+    'FileTypeError', 'XMPParser', 'XMPSerializer'
+]
