@@ -1,4 +1,4 @@
-# Copyright (c) 2021, Ora Lassila & So Many Aircraft
+# Copyright (c) 2021-2022, Ora Lassila & So Many Aircraft
 # All rights reserved.
 #
 # See LICENSE for licensing information
@@ -10,7 +10,7 @@
 # modified because RDFLib did not provide suitable extension points. That code is
 # Copyright (c) 2002-2020, RDFLib Team and is distributed under a similar 3-clause BSD
 # License; see this file: https://github.com/RDFLib/rdflib/blob/master/LICENSE
-
+import rdflib
 from rdflib import RDF, URIRef, BNode, Literal, Namespace, Graph, plugin
 from rdflib.exceptions import Error
 from rdflib.parser import Parser
@@ -23,7 +23,7 @@ import re
 from datetime import datetime
 import os.path
 import pathlib
-from urllib.parse import urlparse
+from urllib.parse import urlparse, unquote_plus
 
 TOOLNAME = "So Many Aircraft xmptools 0.1"
 XMP_NS = "adobe:ns:meta/"
@@ -39,6 +39,16 @@ class FileTypeError(Error):
         self.path = path
         self.original = original
 
+LI_MATCH_PATTERN = re.compile(str(RDF) + "_([0-9]+)")
+LI_CREATE_PATTERN = str(RDF) + "_{0}"
+
+def isContainerItemPredicate(uri):
+    match = LI_MATCH_PATTERN.match(uri)
+    return int(match.group(1)) if match else None
+
+def makeContainerItemPredicate(index):
+    return LI_CREATE_PATTERN.format(index)
+
 class XMPHandler(RDFXMLHandler):
     def reset(self):
         super().reset()
@@ -122,27 +132,27 @@ XMP = Namespace("http://ns.adobe.com/xap/1.0/")
 EXIF = Namespace("http://ns.adobe.com/exif/1.0/")
 CRS = Namespace("http://ns.adobe.com/camera-raw-settings/1.0/")
 DC = Namespace("http://purl.org/dc/elements/1.1/")
+DCT = Namespace("http://purl.org/dc/terms/")
+PHOTOSHOP = Namespace("http://ns.adobe.com/photoshop/1.0/")
 
 class MoreThanOneValue(Error):
     pass
 
-LI_MATCH_PATTERN = re.compile(str(RDF) + "_([0-9]+)")
-LI_CREATE_PATTERN = str(RDF) + "_{0}"
+def URI(u):
+    return u if isinstance(u, URIRef) or u is None else URIRef(u)
 
 def makeFileURI(path):
-    return pathlib.Path(os.path.abspath(path)).as_uri()
+    return URIRef(pathlib.Path(os.path.abspath(path)).as_uri())
 
-def isContainerItemPredicate(uri):
-    match = LI_MATCH_PATTERN.match(uri)
-    return int(match.group(1)) if match else None
-
-def makeContainerItemPredicate(index):
-    return LI_CREATE_PATTERN.format(index)
+def makeFilePath(uri, scheme="file"):
+    u = urlparse(uri)
+    if u.scheme == scheme:
+        return unquote_plus(u.path)
+    else:
+        raise ValueError("URI does not have '{}' scheme: {}".format(scheme, str(uri)))
 
 def getContainerStatements(graph, source, predicate):
-    if not isinstance(source, URIRef):
-        source = URIRef(source)
-    containers = list(graph.objects(source, predicate))
+    containers = list(graph.objects(URI(source), predicate))
     n = len(containers)
     if n == 1:
         statements = graph.triples((containers[0], None, None))
@@ -152,23 +162,52 @@ def getContainerStatements(graph, source, predicate):
     else:
         raise MoreThanOneValue("Expected only one value for {0}".format(predicate))
 
+def adjustTriple(triple, node_map, check_predicates=False):
+    s, p, o = triple
+    for node in node_map:
+        if s == node:
+            s = node_map[node]
+        if check_predicates and p == node:
+            p = node_map[node]
+        if o == node:
+            o = node_map[node]
+    return s, p, o
+
 def adjustNodes(node_map, source_graph, destination_graph=None, check_predicates=False):
     if destination_graph is None:
         destination_graph = type(source_graph)()
-    for s, p, o in source_graph:
-        for node in node_map:
-            if s == node:
-                s = node_map[node]
-            if check_predicates and p == node:
-                p = node_map[node]
-            if o == node:
-                o = node_map[node]
-        destination_graph.add((s, p, o))
+    for triple in source_graph:
+        destination_graph.add(adjustTriple(triple, node_map, check_predicates=check_predicates))
     if isinstance(destination_graph, XMPMetadata):
         destination_graph.url = node_map[source_graph.url]
         destination_graph.sourceIsXMP = False
     return destination_graph
 
+def adjustNodesInPlace(node_map, graph, check_predicates=False):
+    for triple in list(graph.triples((None, None, None))):
+        new_triple = adjustTriple(triple, node_map, check_predicates=check_predicates)
+        if new_triple != triple:
+            graph.add(new_triple)
+            graph.remove(triple)
+    if isinstance(graph, XMPMetadata):
+        graph.url = node_map[graph.url]
+    return graph
+
+def identity(x):
+    return x
+
+def cbd(source, target, resource):
+    # Does not support reified statements (yet)
+    subjects = list([resource])
+    while subjects:
+        subject = subjects.pop()
+        for triple in source.triples((subject, None, None)):
+            target.add(triple)
+            leaf = triple[2]
+            if isinstance(leaf, rdflib.BNode):
+                subjects.append(leaf)
+    return target
+
 JPEG_EXTENSIONS = [".jpg", ".jpeg", ".JPG", ".JPEG"]
 TIFF_EXTENSIONS = [".tif", ".tiff", ".TIF", ".TIFF"]
 DNG_EXTENSIONS  = [".dng", ".DNG"]
@@ -183,7 +222,7 @@ class XMPMetadata(Graph):
             self.sourceIsXMP = False  # signals that content didn't come from an XMP sidecar file
             self.segment_count = 0
         else:
-            self.url = URIRef(url)
+            self.url = URI(url)
             self.sourceIsXMP = True
             self.segment_count = 1
         super().__init__(identifier=self.url)
@@ -244,7 +283,7 @@ class XMPMetadata(Graph):
                 return xmp
             if xmp is None:
                 xmp = cls()  # empty graph
-                xmp.url = URIRef(makeFileURI(os.path.abspath(path)))
+                xmp.url = makeFileURI(os.path.abspath(path))
             xmp.parse(io.BytesIO(stuff[i:j]), format="xmp", publicID=makeFileURI(path))
             xmp.initialized = True
             xmp.segment_count += 1
@@ -265,14 +304,15 @@ class XMPMetadata(Graph):
             elif not isinstance(destination, io.IOBase):
                 raise ValueError("Unsupported destination: " + destination)
         elif not self.sourceIsXMP:
-            raise NotImplementedError("Cannot write metadata back to " + self.url)
+            raise NotImplementedError("Cannot write metadata back to " + str(self.url))
         else:
             destination = self.url
         self.serialize(destination=destination, format="xmp", xmpfile=self.url)
 
     def getOneStatement(self, predicate):
-        statements = list(self.triples((self.url, predicate, None)))
-        return statements[0] if statements and len(statements) > 0 else None
+        for statement in self.triples((self.url, predicate, None)):
+            return statement
+        return None
 
     def getValue(self, predicate):
         statement = self.getOneStatement(predicate)
@@ -291,7 +331,7 @@ class XMPMetadata(Graph):
 
     def getContainerItems(self, predicate):
         statements = getContainerStatements(self, self.url, predicate)
-        return [statement[2].value for statement in statements] if statements else None
+        return [statement[2] for statement in statements] if statements else None
 
     def setContainerItems(self, predicate, values, newtype=RDF.Seq):
         # Having to write code like this is a clear indication that triples are the wrong
@@ -308,7 +348,7 @@ class XMPMetadata(Graph):
                 self.add((container, RDF.type, newtype))
             i = 1
             for value in values:
-                self.add((container, URIRef(makeContainerItemPredicate(i)), Literal(value)))
+                self.add((container, URIRef(makeContainerItemPredicate(i)), value))
                 i += 1
         else:
             statement = self.getOneStatement(predicate)
@@ -316,23 +356,30 @@ class XMPMetadata(Graph):
             self.remove((self.url, predicate, container))
             self.remove((container, None, None))
 
-    def container2repeated(self, predicate):
+    def container2repeated(self, predicate,
+                           new_predicate=None, value_mapper=identity):
         # We understand that this messes with ordering
+        if new_predicate is None:
+            new_predicate = predicate
         items = self.getContainerItems(predicate)
         if items:
             self.setContainerItems(predicate, [])
             for item in items:
-                self.add((self.url, predicate, item))
+                self.add((self.url, new_predicate, value_mapper(item)))
 
-    def repeated2container(self, predicate, newtype=RDF.Seq):
+    def repeated2container(self, predicate,
+                           new_predicate=None, newtype=RDF.Seq, value_mapper=identity):
         # We understand that this messes with ordering
+        if new_predicate is None:
+            new_predicate = predicate
         statements = self.triples((self.url, predicate, None))
         if statements:
             for statement in statements:
                 self.remove(statement)
-            self.setContainerItems(predicate, [o for _, _, o in statements], newtype=newtype)
+            self.setContainerItems(new_predicate, [value_mapper(o) for _, _, o in statements],
+                                   newtype=newtype)
 
-    def adjustImageURI(self, new_extension=None):
+    def adjustImageURI(self, new_extension=None, destination_graph=None, uri_mapper=None):
         # Logic:
         #   1. If no new extension specified, use crs:RawFileName value for the new extension
         #      - if no RawFileName found, signal an error
@@ -349,4 +396,14 @@ class XMPMetadata(Graph):
         else:
             base, _ = os.path.splitext(str(self.url))
             new_uri = base + new_extension
-        return adjustNodes({self.url: URIRef(new_uri)}, self)
+        if uri_mapper:
+            new_uri = uri_mapper(new_uri)
+        node_map = {self.url: URIRef(new_uri)}
+        if destination_graph is self:
+            return adjustNodesInPlace(node_map, self)
+        else:
+            return adjustNodes(node_map, self, destination_graph=destination_graph)
+
+    def cbd(self, resource=None, target=None):
+        # Does not support reified statements (yet)
+        return cbd(self, target or Graph(), resource or self.url)
